/*
	ArchiveSMTP Milter v1.1.b1
	--------------------------
	Copyright (c) 2008, 2009, 2010 Dancing Fortune Software.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions
	are met:

	1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.

	2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
	POSSIBILITY OF SUCH DAMAGE.
 */

#include "ArchiveSMTP.h"

/* GetNextActionSlot: Iterates through a rules action list and returns a newly
 allocated action struct from the end of the list. */
Action* GetNextActionSlot(Rule* r)
{
	Action *ptr, *newAction;
	
	/* Iterate to end of current action list */
	for (newAction = ptr = r->action; ptr != NULL; ptr = ptr->next)
		newAction = ptr;
	
	/* Allocate new action at the end of list */
	ptr = newAction;
	newAction = malloc(sizeof(Action));
	if (ptr)
		ptr->next = newAction;
	else
		r->action = newAction;
	
	return (newAction);
}

/* PrintMail: Prints the full mail data to the given stream. */
void PrintMail(FILE* output, time_t* stamp, mlfiPriv_t *priv)
{
	/* Write mbox header */
	fwrite("From ", sizeof(char), strlen("From "), output);
	if (priv->mlfi_useraddr == NULL) {
		/* We might want to do something about this, but currently ignore*/
	} else {
		fwrite("<", sizeof(char), strlen("<"), output);
		fwrite(priv->mlfi_useraddr, sizeof(char), strlen(priv->mlfi_useraddr), output);
		fwrite(">", sizeof(char), strlen(">"), output);
	}
	fwrite("  ", sizeof(char), strlen("  "), output);
	fwrite(ctime(stamp), sizeof(char), strlen(ctime(stamp)), output);
	
	/* Write Headers to File */
	headerList_t *ptrH = priv->mlfi_header_list;
	if (ptrH != NULL) {
		do {
			fwrite(ptrH->headerName, sizeof(char), strlen(ptrH->headerName), output);
			fwrite(": ", sizeof(char), strlen(": "), output);
			fwrite(ptrH->headerValue, sizeof(char), strlen(ptrH->headerValue), output);
			fwrite("\n", sizeof(char), strlen("\n"), output);
			ptrH = ptrH->next;
		} while (ptrH != NULL);
	}
	else {
		/* Having no headers should be impossible */
		fprintf(stderr, "ArchiveSMTP(PrintMail): WARNING - Message contains no headers.\n");
	}
	
	/* Blank line after headers */
	fwrite("\n", sizeof(char), strlen("\n"), output);
	
	/* Write Body to File */
	bodyChunks_t *ptrB = priv->mlfi_body_chunks;
	if (ptrB != NULL) {
		do {
			fwrite(ptrB->body, sizeof(char), ptrB->size, output);
			ptrB = ptrB->next;
		} while (ptrB != NULL);
	}
	
	/* Blank line after body */
	fwrite("\n\n", sizeof(char), strlen("\n\n"), output);
}

#define CheckTokenMatch(TESTSTR, SRCSTR, OFFSET, RESULT, ITR) RESULT = TRUE; for (ITR = 0; TESTSTR[ITR]; ITR++) { if (SRCSTR[OFFSET+ITR] != TESTSTR[ITR]) { RESULT = FALSE; break; } }
/* PreProcessTokens: Takes a rule action argument string and splits it into
 a linked list of Tokens (if any are present), returning a pointer to the start
 of the list. If no tokens are found a single token will be returned with the
 entire arg string in it and type set to kTokenNone. */
Token* PreProcessTokens(char* args, Rule* r, Action* action)
{
	Token *token, *tPrev = NULL, *tHead = NULL;
	bool match;
	char *strStart = args, *tokenStart;
	int i, tLen = 0, sLen = 0, tType = kTokenNone;
	
	/* Step through all the token delimiters '%' */
	tokenStart = strchr(strStart, '%');
	while (tokenStart) {
		/* Store the string length prior to start of token */
		if (sLen == 0)
			sLen = strcspn(strStart, "%");
		
		/* See if the token is valid */
		switch (tokenStart[1]) {
			case '%':							/* Token is a literal % sign */
				tLen = 2;
				tType = kTokenNone;
				break;
			case 'S':	/* Token starts with SENDER? */
				CheckTokenMatch("ENDER", tokenStart, 2, match, i);
				if (match) {
					CheckTokenMatch("%", tokenStart, 7, match, i);
					if (match) {				/* Token is SENDER */
						tLen = 8;
						tType = kTokenSender;
						break;
					}
					CheckTokenMatch(".USER%", tokenStart, 7, match, i);
					if (match) {				/* Token is SENDER.USER */
						tLen = 13;
						tType = kTokenSenderUser;
						break;
					}
					CheckTokenMatch(".DOMAIN%", tokenStart, 7, match, i);
					if (match) {				/* Token is SENDER.DOMAIN */
						tLen = 15;
						tType = kTokenSenderDomain;
						break;
					}
				}								/* Token is invalid */
				tLen = 0;
				sLen += strcspn(tokenStart+1, "%") + 1;
				tType = kTokenNone;
				break;
			case 'F':	/* Token starts with FROM? */
				CheckTokenMatch("ROM", tokenStart, 2, match, i);
				if (match) {
					CheckTokenMatch("%", tokenStart, 5, match, i);
					if (match) {				/* Token is FROM */
						tLen = 6;
						tType = kTokenFrom;
						break;
					}
					CheckTokenMatch(".USER%", tokenStart, 5, match, i);
					if (match) {				/* Token is FROM.USER */
						tLen = 11;
						tType = kTokenFromUser;
						break;
					}
					CheckTokenMatch(".DOMAIN%", tokenStart, 5, match, i);
					if (match) {				/* Token is FROM.DOMAIN */
						tLen = 13;
						tType = kTokenFromDomain;
						break;
					}
				}								/* Token is invalid */
				tLen = 0;
				sLen += strcspn(tokenStart+1, "%") + 1;
				tType = kTokenNone;
				break;
			case 'A':	/* Token starts with ALLTO? */
				CheckTokenMatch("LLTO", tokenStart, 2, match, i);
				if (match) {
					CheckTokenMatch("%", tokenStart, 6, match, i);
					if (match) {				/* Token is TO */
						tLen = 7;
						tType = kTokenAllTo;
						r->flags |= kRuleFlagUsesToAll;
						action->allTo = TRUE;
						break;
					}
					CheckTokenMatch(".USER%", tokenStart, 6, match, i);
					if (match) {				/* Token is TO.USER */
						tLen = 12;
						tType = kTokenAllToUser;
						r->flags |= kRuleFlagUsesToAll;
						action->allTo = TRUE;
						break;
					}
					CheckTokenMatch(".DOMAIN%", tokenStart, 6, match, i);
					if (match) {				/* Token is TO.DOMAIN */
						tLen = 14;
						tType = kTokenAllToDomain;
						r->flags |= kRuleFlagUsesToAll;
						action->allTo = TRUE;
						break;
					}
				}								/* Token is invalid */
				tLen = 0;
				sLen += strcspn(tokenStart+1, "%") + 1;
				tType = kTokenNone;
				break;
			case 'T':	/* Token starts with TO? */
				CheckTokenMatch("O", tokenStart, 2, match, i);
				if (match) {
					CheckTokenMatch("%", tokenStart, 3, match, i);
					if (match) {				/* Token is TO */
						tLen = 4;
						tType = kTokenTo;
						r->flags |= kRuleFlagUsesTo;
						break;
					}
					CheckTokenMatch(".USER%", tokenStart, 3, match, i);
					if (match) {				/* Token is TO.USER */
						tLen = 9;
						tType = kTokenToUser;
						r->flags |= kRuleFlagUsesTo;
						break;
					}
					CheckTokenMatch(".DOMAIN%", tokenStart, 3, match, i);
					if (match) {				/* Token is TO.DOMAIN */
						tLen = 11;
						tType = kTokenToDomain;
						r->flags |= kRuleFlagUsesTo;
						break;
					}
				}								/* Token is invalid */
				tLen = 0;
				sLen += strcspn(tokenStart+1, "%") + 1;
				tType = kTokenNone;
				break;
			default:							/* Token is invalid */
				tLen = 0;
				sLen += strcspn(tokenStart+1, "%") + 1;
				tType = kTokenNone;
				break;
		}
		
		/* Create token list item if it's valid */
		if (tLen > 0 || tType != kTokenNone) {
			token = malloc(sizeof(Token));
			token->field = tType;
			if (sLen > 0) {
				token->preText = malloc(sizeof(char) * (sLen + 1));
				strncpy(token->preText, strStart, sLen);
				token->preText[sLen] = '\0';
			} else {
				token->preText = NULL;
			}
			token->next = NULL;
			if (!tHead)
				tHead = token;
			if (tPrev)
				tPrev->next = token;
			tPrev = token;
			
			/* Update pointers */
			sLen = 0;
			strStart = tokenStart+tLen;
			tokenStart = strchr(strStart, '%');
		}
		else {
			/* Treat invalid token as regular string and continue parsing after */
			tokenStart = strchr(tokenStart+1, '%');
		}
	}
	
	/* Add any text found after final token to token list (if no tokens exist
	 this will be the only action taken) */
	token = malloc(sizeof(Token));
	token->field = kTokenNone;
	token->preText = malloc(sizeof(char) * (strlen(strStart)+1));
	strcpy(token->preText, strStart);
	token->next = NULL;
	if (!tHead)
		tHead = token;
	if (tPrev)
		tPrev->next = token;
	
	/* Return start of list */
	return (tHead);
}

/* GetPreTokenStrLen: Returns the string length of a token list prior to subbing
 in tokens, including space for the null terminator. */
int GetPreTokenLen(Token* tHead)
{
	Token *itr;
	int len = 0;
	
	/* Iterate through token list getting preText string lengths */
	itr = tHead;
	while (itr) {
		if (itr->preText)
			len += strlen(itr->preText);
		itr = itr->next;
	}
	
	return (len+1);
}

/* SubTokens: Subsitutes tokens in a string with their real values, returning
 a new string. */
char* SubTokens(Action* action, mlfiPriv_t *priv)
{
	Token *itr = action->args;
	char *tokenSub;
	int sLen = 0, senderLen, senderULen, fromLen, fromULen, toLen, toULen, allToLen, allToULen;
	
	/* Store length of possible substitution strings */
	if (priv->mlfi_username != NULL) {
		senderLen = strlen(priv->mlfi_username);
		senderULen = strcspn(priv->mlfi_username, "@");
	} else
		senderLen = senderULen = 0;
	fromLen = strlen(priv->mlfi_useraddr);
	fromULen = strcspn(priv->mlfi_useraddr, "@");
	if (priv->mlfi_rcptaddr != NULL) {
		toLen = strlen(priv->mlfi_rcptaddr);
		toULen = strcspn(priv->mlfi_rcptaddr, "@");
	} else
		toLen = toULen = 0;
	if (action->addrItr != NULL) {
		allToLen = strlen(action->addrItr->address);
		allToULen = strcspn(action->addrItr->address, "@");
	} else
		allToLen = allToULen = 0;
	
	/* Calculate length of all substitutions */
	while (itr) {
		switch(itr->field) {
			case kTokenSender:
				sLen += senderLen;
				break;
			case kTokenSenderUser:
				sLen += senderULen;
				break;
			case kTokenSenderDomain:
				sLen += (senderLen - senderULen);
				break;
			case kTokenFrom:
				sLen += fromLen;
				break;
			case kTokenFromUser:
				sLen += fromULen;
				break;
			case kTokenFromDomain:
				sLen += (fromLen - fromULen);
				break;
			case kTokenTo:
				sLen += toLen;
				break;
			case kTokenToUser:
				sLen += toULen;
				break;
			case kTokenToDomain:
				sLen += (toLen - toULen);
				break;
			case kTokenAllTo:
				sLen += allToLen;
				break;
			case kTokenAllToUser:
				sLen += allToULen;
				break;
			case kTokenAllToDomain:
				sLen += (allToLen - allToULen);
				break;
			default:
				break;
		}
		itr = itr->next;
	}
	
	/* Allocate space to hold the full path and fill it in */
	tokenSub = malloc(sizeof(char) * (action->pathLen + sLen + 1));
	tokenSub[0] = '\0';
	itr = action->args;
	while (itr) {
		if (itr->preText)
			strcat(tokenSub, itr->preText);
		
		switch(itr->field) {
			case kTokenNone:
				break;
			case kTokenSender:
				if (senderLen > 0)
					strcat(tokenSub, priv->mlfi_username);
				break;
			case kTokenSenderUser:
				if (senderULen > 0)
					strncat(tokenSub, priv->mlfi_username, senderULen);
				break;
			case kTokenSenderDomain:
				if ((senderLen - (senderULen+1)) > 0)
					strcat(tokenSub, priv->mlfi_username+senderULen+1);
				break;
			case kTokenFrom:
				if (fromLen > 0)
					strcat(tokenSub, priv->mlfi_useraddr);
				break;
			case kTokenFromUser:
				if (fromULen > 0)
					strncat(tokenSub, priv->mlfi_useraddr, fromULen);
				break;
			case kTokenFromDomain:
				if ((fromLen - (fromULen+1)) > 0)
				strcat(tokenSub, priv->mlfi_useraddr+fromULen+1);
				break;
			case kTokenTo:
				if (toLen > 0)
					strcat(tokenSub, priv->mlfi_rcptaddr);
				break;
			case kTokenToUser:
				if (toULen > 0)
					strncat(tokenSub, priv->mlfi_rcptaddr, toULen);
				break;
			case kTokenToDomain:
				if ((toLen - (toULen+1)) > 0)
					strcat(tokenSub, priv->mlfi_rcptaddr+toULen+1);
				break;
			case kTokenAllTo:
				if (allToLen > 0)
					strcat(tokenSub, action->addrItr->address);
				break;
			case kTokenAllToUser:
				if (allToULen > 0)
					strncat(tokenSub, action->addrItr->address, allToULen);
				break;
			case kTokenAllToDomain:
				if ((allToLen - (allToULen+1)) > 0)
					strcat(tokenSub, action->addrItr->address+allToULen+1);
				break;
			default:
				break;
		}
		itr = itr->next;
	}
	
	/* Return new string with tokens subbed */
	return (tokenSub);
}

/* ActionStoreIn: Stores the mail to the path specified in the action using mbox
 format. */
bool ActionStoreIn(Action* storein, time_t* stamp, mlfiPriv_t *priv)
{
	if (storein->allTo && storein->addrItr == NULL) {
		/* We need to perform the action for every matched TO address, this is
		 the first iteration so we need to set our iterator to the first matched
		 address, if any. */
		storein->addrItr = priv->mlfi_toAddr_list;
		while (storein->addrItr != NULL && !storein->addrItr->matched)
			storein->addrItr = storein->addrItr->next;
	}
		
	/* Open the correct file */
	char *path = SubTokens(storein, priv);
	if (storein->allTo)
		VerbosePrint(1, "%s, ", path);
	
	/* Open File Descriptor */
	int outFD = open(path, O_WRONLY|O_CREAT, S_IRUSR|S_IWUSR|S_IWGRP);
	if (outFD == -1) {
		/* We couldn't open the file, so call the clean up routine */
		char msg[strlen(path)+24];
		sprintf(msg, "Could not open file '%s'.", path);
		MemErrorPrint("mlfi_eom", msg, 2);
		return (FALSE);
	}
	
	/* Obtain lock on current EOF to num bytes past that we need to write */
	struct flock opts;
	opts.l_start    = 0;        /* starting offset */
	opts.l_len      = priv->mlfi_bytes; /* len = 0 means until end of file */
	opts.l_type     = F_WRLCK;  /* lock type: read/write, etc. */
	opts.l_whence   = SEEK_END; /* type of l_start */
	
	/* Keep trying to get lock while error val is: locked (EAGAIN) */
	while (fcntl(outFD, F_SETLK, &opts) == -1) {
		if (errno != EAGAIN) {
			/* There was a dead lock or some other error, call clean up */
			char msg[strlen(path)+32];
			sprintf(msg, "Could not get write lock on '%s'.", path);
			MemErrorPrint("mlfi_eom", msg, 2);
			return (FALSE);
		}
	}
	
	/* Connect stream to FD */
	FILE *output = fdopen(outFD, "a");
	if (output == NULL) {
		/* We failed to connect the stream, so call clean up */
		MemErrorPrint("mlfi_eom", "Could not connect stream to FD.", 2);
		return (FALSE);
	}
	
	/* Write mail to the stream */
	PrintMail(output, stamp, priv);
	
	/* Unlock file area */
	if (fcntl(outFD, F_UNLCK, &opts) == -1) {
		fprintf(stderr, "ArchiveSMTP(mlfi_eom): WARNING - Could not release write lock. (errno: %d:%s)\n", errno, strerror(errno));
	}
	
	/* Close stream (will also close fd) */
	fclose(output);
	
	/* Free dynamically generated path string */
	free(path);
	
	if (storein->allTo && storein->addrItr != NULL) {
		/* In all to mode, find next matched address if any */
		storein->addrItr = storein->addrItr->next;
		while (storein->addrItr != NULL && !storein->addrItr->matched)
			storein->addrItr = storein->addrItr->next;
		
		if (storein->addrItr)
			return (ActionStoreIn(storein, stamp, priv));
	}
		
	return (TRUE);
}

/* ActionAddHeader: Adds a header field to the mail by splitting the action arg
 at ':' to get the header name and value. */
bool ActionAddHeader(Action* addheader, SMFICTX* ctx, mlfiPriv_t *priv)
{
	/* Check the MTA supports us doing this */
	if (!addHeadersSupported) {
		VerbosePrint(2, "%s ", "Adding headers NOT supported by MTA - Ignoring action");
		return (TRUE);
	}
	
	if (addheader->allTo && addheader->addrItr == NULL) {
		/* We need to perform the action for every matched TO address, this is
		 the first iteration so we need to set our iterator to the first matched
		 address, if any. */
		addheader->addrItr = priv->mlfi_toAddr_list;
		while (addheader->addrItr != NULL && !addheader->addrItr->matched)
			addheader->addrItr = addheader->addrItr->next;
	}
	
	bool result = TRUE;
	char *name, *lName, *val, *header = SubTokens(addheader, priv);
	
	/* First split the header into its name and value pair */
	name = strtok(header, ":");
	val = strtok(NULL, "");
	VerbosePrint(2, "Name: %s ", name);
	VerbosePrint(2, "Val: %s ", val);
	
	/* Determin the type of header to add */
	if (val) {
		lName = malloc(strlen(name)+1);
		strcpy(lName, name);
		StringToLower(lName);
		
		/* If the header adds a recipient we need to call an extra method */
		if (!strcmp(lName, "to") ||
			!strcmp(lName, "cc") ||
			!strcmp(lName, "bcc")) {
			VerbosePrint(2, "%s ", "Adding recipient");
			result = smfi_addrcpt(ctx, val) == MI_SUCCESS;
		}
		free(lName);
		/* Add the header with value */
		result |= smfi_addheader(ctx, name, val);
	} else {
		/* Add the header with null value */
		result = smfi_addheader(ctx, name, "\0") == MI_SUCCESS;
	}
	
	/* Free dynamically generated header string */
	free(header);
	
	if (addheader->allTo && addheader->addrItr != NULL) {
		/* In all to mode, find next matched address if any */
		addheader->addrItr = addheader->addrItr->next;
		while (addheader->addrItr != NULL && !addheader->addrItr->matched)
			addheader->addrItr = addheader->addrItr->next;
		
		if (addheader->addrItr)
			return (ActionAddHeader(addheader, ctx, priv));
	}
	
	return (result);
}

/* ActionPipe: Pipes the full mail to the program specified in arg. */
bool ActionPipe(Action* pipeto, time_t* stamp, mlfiPriv_t* priv)
{
	if (pipeto->allTo && pipeto->addrItr == NULL) {
		/* We need to perform the action for every matched TO address, this is
		 the first iteration so we need to set our iterator to the first matched
		 address, if any. */
		pipeto->addrItr = priv->mlfi_toAddr_list;
		while (pipeto->addrItr != NULL && !pipeto->addrItr->matched)
			pipeto->addrItr = pipeto->addrItr->next;
	}
	
	char *program = SubTokens(pipeto, priv);
	
	/* Try to connect the pipe to given program */
	FILE *pipe = popen(program, "w");
	if (pipe == NULL) {
		/* We couldn't open the program */
		char msg[strlen(program)+27]; // 27 bytes is for the message text below
		sprintf(msg, "Could not open program '%s'.", program);
		MemErrorPrint("ActionPipe", msg, 2);
		return (FALSE);
	}
	
	/* Print the full mbox formatted mail to the pipe */
	PrintMail(pipe, stamp, priv);
	
	/* Close pipe and return */
	pclose(pipe);
	
	/* Free dynamically generated program name */
	free(program);
	
	if (pipeto->allTo && pipeto->addrItr != NULL) {
		/* In all to mode, find next matched address if any */
		pipeto->addrItr = pipeto->addrItr->next;
		while (pipeto->addrItr != NULL && !pipeto->addrItr->matched)
			pipeto->addrItr = pipeto->addrItr->next;
		
		if (pipeto->addrItr)
			return (ActionPipe(pipeto, stamp, priv));
	}
	
	return (TRUE);
}
