/*
    ArchiveSMTP Milter v1.1.b1
    --------------------------
	Copyright (c) 2008, 2009, 2010 Dancing Fortune Software.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions
	are met:
 
	1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
 
	2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
	POSSIBILITY OF SUCH DAMAGE.
 */
 
typedef struct Condition_s Condition;
/*
	Condition_s
	
	This lives above rules on the tree
	
		 Test   Test
			\    /
	Test  Condition
	 \    /
	Condition
		|
		|
	  Rule
	
	A Condition describes an OR or AND performed on a combination of tests, other conditions
	or nothing. It consists of two pointers, each to either a test or a condition. These are
	evaluated to determine the status of the condition.
	
	The function in question is EvaluateCondition(Condition*)
	
	A Condition can be initialised with InitialiseCondition(Condition*)
 */
struct Condition_s
{
	int			type;
	Condition	*p1, *p2;
	Test		*t1, *t2; /* Quirk: the initial design called for it, but
							  in practice *t2 isn't ever actually used. */
	bool		negate;   /* due to implementation, only the first branch is ever negated */
};

enum
{
	kConditionTypeUndefined, /*  condition type is not yet defined and
								testing it will produce an error	  */
	kConditionTypeOr,		 /*  condition is an OR of two other
								conditions or tests					  */
	kConditionTypeAnd		 /*  condition is an AND of two other
								conditions or tests					  */
};

