/*
	ArchiveSMTP Milter v1.1.b1
	--------------------------
	Copyright (c) 2008, 2009, 2010 Dancing Fortune Software.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions
	are met:

	1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.

	2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
	POSSIBILITY OF SUCH DAMAGE.
 */

#include "ArchiveSMTP.h"

/*
 EvaluateRule's purpose is mainly to hide the implementation detail
 of evaluating the rule's truth or not, even though that isn't super
 complicated.
 */
int EvaluateRule(Rule* r, mlfiPriv_t *priv)
{
	priv->currentRule = r; /* set the current rule so the tests can look up its flags */
	
	/* reset the address matches for token use if necessary */
	if (r->flags & (kRuleFlagUsesTo | kRuleFlagUsesToAll))
	{
		priv->mlfi_rcptaddr = NULL;
		toAddrList_t *a;
		for (a = priv->mlfi_toAddr_list; a; a = a->next)
			a->matched = FALSE;
	}
	return EvaluateCondition(r->root, priv);
}

/* allocate a new rule. The rule is a pristine slate ready for writing on. */
Rule* CreateNewRule()
{
	Rule* r = malloc(sizeof(Rule));
	r->root = NULL;
	r->action = NULL;
	r->halt = FALSE;
	return r;
}

void DeleteRule(Rule* r)
{
	Action *del, *ptr = r->action;
	while (ptr) {
		del = ptr;
		ptr = ptr->next;
		if (del->args)
			free(del->args);
		free(del);
	}
	
	DeleteCondition(r->root);
}

/* AddRuleToRuleList: Adds a new rule to our list of rules to process. */
void AddRuleToRuleList(Rule* newRule)
{
	// Iterate to end of rule list, last rule will be at p
	Rule *i = firstRule;
	Rule *p = NULL;
	while (i) {
		p = i;
		i = i->next;
	}
	
	// Add the new rule to the end of the rule list
	newRule->next = NULL;
	if (p)
		p->next = newRule;
	else
		firstRule = newRule;
}
