/*
    ArchiveSMTP Milter v1.1.b1
    --------------------------
	Copyright (c) 2008, 2009, 2010 Dancing Fortune Software.
	All rights reserved.

	Redistribution and use in source and binary forms, with or without
	modification, are permitted provided that the following conditions
	are met:
 
	1. Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
 
	2. Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in the
	documentation and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
	IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
	ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
	LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
	CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
	SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
	INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
	CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
	POSSIBILITY OF SUCH DAMAGE.
 */

#include "ArchiveSMTP.h"

/*
	PrintTest(Test*)
	
	Prints the contents of a test to standard output. Used for debugging
	one's config file.
*/
void PrintTest(Test* t)
{
	switch (t->fieldToTest)
	{
		case kTestSMTPSender:
			printf("SMTP Sender ");
			break;
		case kTestFromAddress:
			printf("From Address ");
			break;
		case kTestToAddress:
			printf("To Address ");
			break;
		case kTestUndefined:
			printf("[Undefined Field] ");
			break;
		case kTestMatchAll:
			printf("[Always True] ");
			return;
		case kTestNoMatch:
			printf("No Rule has been matched yet ");
			return;
	}
	
	if (t->negate)
		printf("!");
	
	switch (t->locationToTest)
	{
		case kLocationFieldStart:
			printf("starts with ");
			break;
		case kLocationFieldEnd:
			printf("ends with ");
			break;
		case kLocationFieldAny:
			printf("contains ");
			break;
		case kLocationFieldExact:
			printf("is ");
			break;
		case kLocationFieldUndefined:
			printf("is operated on in some way by ");
			break;
	}
	
	printf("\"%s\"", t->textToFind);
}

/* PerformTest: Run the test on the current data and return the result. */
int PerformTest(Test* t, const char* testData)
{
	int result = FALSE;
		
	/* See what part of the data we are testing and test it */
	if (testData)
	{
		VerbosePrint(1, "\"%s\" ", testData);
		/* Check length, if test data is less than find text no need to test */
		int lengthOfTextToFind = strlen(t->textToFind);
		int lengthOfStringToTest = strlen(testData);
		if (lengthOfStringToTest >= lengthOfTextToFind) {
			int i,j;
			if (t->negate) {
				VerbosePrint(1, "%s", "!");
			}
			switch (t->locationToTest) {
				case kLocationFieldStart:		/* Find text at start of data */
					VerbosePrint(1, "%s ", "starts");
					for (i = 0; i < lengthOfTextToFind; i++) {
						if (t->textToFind[i] != testData[i])
						{
							result = FALSE;
							break;
						} else {
							result = TRUE;
						}
					}
					break;
				case kLocationFieldEnd:			/* Find text at end of data */
					VerbosePrint(1, "%s ", "ends");
					for (i = lengthOfTextToFind-1, j = lengthOfStringToTest-1; i >= 0; i--, j--) {
						if (t->textToFind[i] != testData[j]) {
							result = FALSE;
							break;
						} else {
							result = TRUE;
						}
					}
					break;
				case kLocationFieldAny:			/* Find text any where in data */
					VerbosePrint(1, "%s ", "contains");
					if (strstr(testData, t->textToFind)) {
						result = TRUE;
					} else {
						result = FALSE;
					}
					break;
				case kLocationFieldExact:
					VerbosePrint(1, "%s ", "is");
					if (!strcmp(testData, t->textToFind))
						result = TRUE;
					else
						result = FALSE;
					break;
				case kLocationFieldUndefined:
				default:
					VerbosePrint(1, "%s ", "confused");
					break;
			}
		}
	} else if (t->textToFind[0] == '\0' || verboseLevel > 0) {
		VerbosePrint(1, "%s ", "[no data]");
		switch (t->locationToTest) {
			case kLocationFieldStart: VerbosePrint(1, "%s ", "starts"); break;
			case kLocationFieldEnd: VerbosePrint(1, "%s ", "ends"); break;
			case kLocationFieldAny: VerbosePrint(1, "%s ", "contains"); break;
			case kLocationFieldExact: VerbosePrint(1, "%s ", "is"); break;
			default: VerbosePrint(1, "%s ", "confused"); break;
		}
		/* If our text to find is null then we want to match no data */
		if (t->textToFind[0] == '\0')
			result = TRUE;
	}
	VerbosePrint(1, "\"%s\" ", t->textToFind)
	
	if (verboseLevel > 0) {
		if (result) {
			if (t->negate) {
				VerbosePrint(1, "%s", "[false]");
			} else {
				VerbosePrint(1, "%s", "[true]");
			}
		} else {
			if (t->negate) {
				VerbosePrint(1, "%s", "[true]");
			} else {
				VerbosePrint(1, "%s", "[false]");
			}
		}
	}
	
	if (t->negate)
		return !result;
	else
		return (result);
}

Test* ParseTest(FILE* file, int testField)
{
	Test*	theTest;
	char	input;
	
	theTest = malloc(sizeof(Test));
	theTest->fieldToTest = testField;
	theTest->negate = FALSE; /* by default, do not negate the result */
	/*
		Currently, the file should be up to a whitespace after the field to test was identified.
		If the test is a "match all" or "no match" test then we want to set some defaults and
		return - they have no parameters, so we don't want to read more. Otherwise, we:
		
		1. eat whitespace
		2. read in the next word, the location in the field
		3. eat more whitespace
		4. read in the text to check
	*/
	
	if (testField == kTestNoMatch || testField == kTestMatchAll)
	{
		theTest->locationToTest = kLocationFieldUndefined;
		theTest->textToFind = NULL;
		return theTest;
	}
	
	input = EatWhitespace(file);
	if (input == '!')
	{
		/* Negation! This flag means the Test result should be inverted. */
		theTest->negate = TRUE;
		input = EatWhitespace(file); /* get the next character, and a space is okay */
	}
	switch (input)
	{
		case 'S':
		case 's':
			/* start */
			CheckStringIsNextInFile(file, "tarts", "Parsing error; expected \"Starts\"\n");
			theTest->locationToTest = kLocationFieldStart;
			break;
		case 'E':
		case 'e':
			/* end */
			CheckStringIsNextInFile(file, "nds", "Parsing error; expected \"Ends\"\n");
			theTest->locationToTest = kLocationFieldEnd;
			break;
		case 'C':
		case 'c':
			/* anywhere */
			CheckStringIsNextInFile(file, "ontains","Parsing error; expected \"Contains\"\n");
			theTest->locationToTest = kLocationFieldAny;
			break;
		case 'I':
		case 'i':
			/* is */
			CheckStringIsNextInFile(file, "s","Parsing error; expected \"Is\"\n");
			theTest->locationToTest = kLocationFieldExact;
			break;
		default:
			fprintf(stderr, "Parsing error; expected a field location\n");
			exit(EX_SOFTWARE);
			break;
	}
	
	/* next item should be the string to search for. We have to read it in. */
	theTest->textToFind = EatWhiteAndReadString(file);
	
	/* convert to all lower-case if its not a username */
	if (testField != kTestSMTPSender)
		StringToLower(theTest->textToFind);
	
	return (theTest);
}

